/*
 * bsc1195950_i915_gem
 *
 * Fix for CVE-2022-0330, bsc#1195950
 * (drivers/gpu/drm/i915/gem/i915_pages.c part)
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_DRM_I915)

#if !IS_MODULE(CONFIG_DRM_I915)
#error "Live patch supports only CONFIG_DRM_I915=m"
#endif

/* klp-ccp: from drivers/gpu/drm/i915/gem/i915_gem_pages.c */
#include "bsc1195950_common.h"

/* klp-ccp: from drivers/gpu/drm/i915/intel_uncore.h */
static void (*klpe_intel_uncore_forcewake_get)(struct intel_uncore *uncore,
				enum forcewake_domains domains);
static void (*klpe_intel_uncore_forcewake_put)(struct intel_uncore *uncore,
				enum forcewake_domains domains);

static int (*klpe___intel_wait_for_register_fw)(struct intel_uncore *uncore,
				 i915_reg_t reg,
				 u32 mask,
				 u32 value,
				 unsigned int fast_timeout_us,
				 unsigned int slow_timeout_ms,
				 u32 *out_value);

/* klp-ccp: from drivers/gpu/drm/i915/intel_runtime_pm.h */
static intel_wakeref_t (*klpe_intel_runtime_pm_get_if_in_use)(struct intel_runtime_pm *rpm);

#define klpr_with_intel_runtime_pm_if_in_use(rpm, wf) \
	for ((wf) = (*klpe_intel_runtime_pm_get_if_in_use)(rpm); (wf);	\
	     klpr_intel_runtime_pm_put((rpm), (wf)), (wf) = 0)

static void (*klpe_intel_runtime_pm_put_unchecked)(struct intel_runtime_pm *rpm);

#if IS_ENABLED(CONFIG_DRM_I915_DEBUG_RUNTIME_PM)
#error "klp-ccp: non-taken branch"
#else
static inline void
klpr_intel_runtime_pm_put(struct intel_runtime_pm *rpm, intel_wakeref_t wref)
{
	(*klpe_intel_runtime_pm_put_unchecked)(rpm);
}
#endif

/* klp-ccp: from drivers/gpu/drm/i915/gem/i915_gem_object.h */
static void (*klpe_i915_gem_object_make_unshrinkable)(struct drm_i915_gem_object *obj);

/* klp-ccp: from drivers/gpu/drm/i915/gem/i915_gem_mman.h */
static void (*klpe_i915_gem_object_release_mmap_offset)(struct drm_i915_gem_object *obj);

/* klp-ccp: from drivers/gpu/drm/i915/gem/i915_gem_pages.c */
static void __i915_gem_object_reset_page_iter(struct drm_i915_gem_object *obj)
{
	struct radix_tree_iter iter;
	void __rcu **slot;

	rcu_read_lock();
	radix_tree_for_each_slot(slot, &obj->mm.get_page.radix, &iter, 0)
		radix_tree_delete(&obj->mm.get_page.radix, iter.index);
	rcu_read_unlock();
}

static void unmap_object(struct drm_i915_gem_object *obj, void *ptr)
{
	if (is_vmalloc_addr(ptr))
		vunmap(ptr);
	else
		kunmap(kmap_to_page(ptr));
}

/* New. */
static struct mutex *klpp_tlb_invalidate_lock;

struct reg_and_bit {
	i915_reg_t reg;
	u32 bit;
};

/* New. */
static struct reg_and_bit
klpp_get_reg_and_bit(const struct intel_engine_cs *engine, const bool gen8,
		const i915_reg_t *regs, const unsigned int num)
{
	const unsigned int class = engine->class;
	struct reg_and_bit rb = { };

	if (WARN_ON_ONCE(class >= num || !regs[class].reg))
		return rb;

	rb.reg = regs[class];
	if (gen8 && class == VIDEO_DECODE_CLASS)
		rb.reg.reg += 4 * engine->instance; /* GEN8_M2TCR */
	else
		rb.bit = engine->instance;

	rb.bit = BIT(rb.bit);

	return rb;
}

/* New. */
static void klpp_intel_gt_invalidate_tlbs(struct intel_gt *gt)
{
	static const i915_reg_t gen8_regs[] = {
		[RENDER_CLASS]			= KLPP_GEN8_RTCR,
		[VIDEO_DECODE_CLASS]		= KLPP_GEN8_M1TCR, /* , GEN8_M2TCR */
		[VIDEO_ENHANCEMENT_CLASS]	= KLPP_GEN8_VTCR,
		[COPY_ENGINE_CLASS]		= KLPP_GEN8_BTCR,
	};
	static const i915_reg_t gen12_regs[] = {
		[RENDER_CLASS]			= KLPP_GEN12_GFX_TLB_INV_CR,
		[VIDEO_DECODE_CLASS]		= KLPP_GEN12_VD_TLB_INV_CR,
		[VIDEO_ENHANCEMENT_CLASS]	= KLPP_GEN12_VE_TLB_INV_CR,
		[COPY_ENGINE_CLASS]		= KLPP_GEN12_BLT_TLB_INV_CR,
	};
	struct drm_i915_private *i915 = gt->i915;
	struct intel_uncore *uncore = gt->uncore;
	struct intel_engine_cs *engine;
	enum intel_engine_id id;
	const i915_reg_t *regs;
	unsigned int num = 0;

	if (I915_SELFTEST_ONLY(gt->awake == -ENODEV))
		return;

	if (INTEL_GEN(i915) == 12) {
		regs = gen12_regs;
		num = ARRAY_SIZE(gen12_regs);
	} else if (INTEL_GEN(i915) >= 8 && INTEL_GEN(i915) <= 11) {
		regs = gen8_regs;
		num = ARRAY_SIZE(gen8_regs);
	} else if (INTEL_GEN(i915) < 8) {
		return;
	}

	if (drm_WARN_ONCE(&i915->drm, !num,
			  "Platform does not implement TLB invalidation!"))
		return;

	GEM_TRACE("\n");

	assert_rpm_wakelock_held(&i915->runtime_pm);

	mutex_lock(klpp_tlb_invalidate_lock);
	(*klpe_intel_uncore_forcewake_get)(uncore, FORCEWAKE_ALL);

	for_each_engine(engine, gt, id) {
		/*
		 * HW architecture suggest typical invalidation time at 40us,
		 * with pessimistic cases up to 100us and a recommendation to
		 * cap at 1ms. We go a bit higher just in case.
		 */
		const unsigned int timeout_us = 100;
		const unsigned int timeout_ms = 4;
		struct reg_and_bit rb;

		rb = klpp_get_reg_and_bit(engine, regs == gen8_regs, regs, num);
		if (!i915_mmio_reg_offset(rb.reg))
			continue;

		intel_uncore_write_fw(uncore, rb.reg, rb.bit);
		if ((*klpe___intel_wait_for_register_fw)(uncore,
						 rb.reg, rb.bit, 0,
						 timeout_us, timeout_ms,
						 NULL))
			drm_err_ratelimited(&gt->i915->drm,
					    "%s TLB invalidation did not complete in %ums!\n",
					    engine->name, timeout_ms);
	}

	(*klpe_intel_uncore_forcewake_put)(uncore, FORCEWAKE_ALL);
	mutex_unlock(klpp_tlb_invalidate_lock);
}

struct sg_table *
klpp___i915_gem_object_unset_pages(struct drm_i915_gem_object *obj)
{
	struct sg_table *pages;

	pages = fetch_and_zero(&obj->mm.pages);
	if (IS_ERR_OR_NULL(pages))
		return pages;

	if (i915_gem_object_is_volatile(obj))
		obj->mm.madv = I915_MADV_WILLNEED;

	(*klpe_i915_gem_object_make_unshrinkable)(obj);

	if (obj->mm.mapping) {
		unmap_object(obj, page_mask_bits(obj->mm.mapping));
		obj->mm.mapping = NULL;
	}

	__i915_gem_object_reset_page_iter(obj);
	obj->mm.page_sizes.phys = obj->mm.page_sizes.sg = 0;

	/*
	 * Fix CVE-2022-0330
	 *  +7 lines
	 */
	if (test_and_clear_bit(KLPP_I915_BO_WAS_BOUND_BIT, &obj->flags)) {
		struct drm_i915_private *i915 = to_i915(obj->base.dev);
		intel_wakeref_t wakeref;

		klpr_with_intel_runtime_pm_if_in_use(&i915->runtime_pm, wakeref)
			klpp_intel_gt_invalidate_tlbs(&i915->gt);
	}

	return pages;
}

int klpp___i915_gem_object_put_pages(struct drm_i915_gem_object *obj)
{
	struct sg_table *pages;
	int err;

	if (i915_gem_object_has_pinned_pages(obj))
		return -EBUSY;

	/* May be called by shrinker from within get_pages() (on another bo) */
	mutex_lock(&obj->mm.lock);
	if (unlikely(atomic_read(&obj->mm.pages_pin_count))) {
		err = -EBUSY;
		goto unlock;
	}

	(*klpe_i915_gem_object_release_mmap_offset)(obj);

	/*
	 * ->put_pages might need to allocate memory for the bit17 swizzle
	 * array, hence protect them from being reaped by removing them from gtt
	 * lists early.
	 */
	pages = klpp___i915_gem_object_unset_pages(obj);

	/*
	 * XXX Temporary hijinx to avoid updating all backends to handle
	 * NULL pages. In the future, when we have more asynchronous
	 * get_pages backends we should be better able to handle the
	 * cancellation of the async task in a more uniform manner.
	 */
	if (!pages && !i915_gem_object_needs_async_cancel(obj))
		pages = ERR_PTR(-EINVAL);

	if (!IS_ERR(pages))
		obj->ops->put_pages(obj, pages);

	err = 0;
unlock:
	mutex_unlock(&obj->mm.lock);

	return err;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/livepatch.h>
#include "livepatch_bsc1195950.h"
#include "../kallsyms_relocs.h"
#include "../shadow.h"

struct klp_bsc1195950_shared_state
{
	unsigned long refcount;
	struct mutex tlb_invalidate_lock;
};

static struct klp_bsc1195950_shared_state *klp_bsc1195950_shared_state;

#define KLP_BSC1195950_SHARED_STATE_ID KLP_SHADOW_ID(1195950, 0)

static int klp_bsc1195950_init_shared_state(void *obj,
					    void *shadow_data,
					    void *ctor_dat)
{
	struct klp_bsc1195950_shared_state *s = shadow_data;

	memset(s, 0, sizeof(*s));
	mutex_init(&s->tlb_invalidate_lock);

	return 0;
}

/* Must be called with module_mutex held. */
static int __klp_bsc1195950_get_shared_state(void)
{
	klp_bsc1195950_shared_state =
		klp_shadow_get_or_alloc(NULL, KLP_BSC1195950_SHARED_STATE_ID,
					sizeof(*klp_bsc1195950_shared_state),
					GFP_KERNEL,
					klp_bsc1195950_init_shared_state, NULL);
	if (!klp_bsc1195950_shared_state)
		return -ENOMEM;

	++klp_bsc1195950_shared_state->refcount;

	klpp_tlb_invalidate_lock =
		&klp_bsc1195950_shared_state->tlb_invalidate_lock;

	return 0;
}

/* Must be called with module_mutex held. */
static void __klp_bsc1195950_put_shared_state(void)
{
	--klp_bsc1195950_shared_state->refcount;

	if (!klp_bsc1195950_shared_state->refcount) {
		mutex_destroy(&klp_bsc1195950_shared_state->tlb_invalidate_lock);
		klp_shadow_free(NULL, KLP_BSC1195950_SHARED_STATE_ID, NULL);
	}

	klp_bsc1195950_shared_state = NULL;
}

#define LIVEPATCHED_MODULE "i915"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "__intel_wait_for_register_fw",
	  (void *)&klpe___intel_wait_for_register_fw, "i915" },
	{ "i915_gem_object_make_unshrinkable",
	  (void *)&klpe_i915_gem_object_make_unshrinkable, "i915" },
	{ "i915_gem_object_release_mmap_offset",
	  (void *)&klpe_i915_gem_object_release_mmap_offset, "i915" },
	{ "intel_runtime_pm_get_if_in_use",
	  (void *)&klpe_intel_runtime_pm_get_if_in_use, "i915" },
	{ "intel_runtime_pm_put_unchecked",
	  (void *)&klpe_intel_runtime_pm_put_unchecked, "i915" },
	{ "intel_uncore_forcewake_get",
	  (void *)&klpe_intel_uncore_forcewake_get, "i915" },
	{ "intel_uncore_forcewake_put",
	  (void *)&klpe_intel_uncore_forcewake_put, "i915" },
};

static int livepatch_bsc1195950_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, LIVEPATCHED_MODULE))
		return 0;

	ret = __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
	WARN(ret, "livepatch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block livepatch_bsc1195950_module_nb = {
	.notifier_call = livepatch_bsc1195950_module_notify,
	.priority = INT_MIN+1,
};

int livepatch_bsc1195950_i915_gem_pages_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	ret = __klp_bsc1195950_get_shared_state();
	if (ret)
		goto out;
	if (find_module(LIVEPATCHED_MODULE)) {
		ret = __klp_resolve_kallsyms_relocs(klp_funcs,
						    ARRAY_SIZE(klp_funcs));
		if (ret) {
			__klp_bsc1195950_put_shared_state();
			goto out;
		}
	}

	ret = register_module_notifier(&livepatch_bsc1195950_module_nb);
	if (ret)
		__klp_bsc1195950_put_shared_state();
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void livepatch_bsc1195950_i915_gem_pages_cleanup(void)
{
	unregister_module_notifier(&livepatch_bsc1195950_module_nb);
	mutex_lock(&module_mutex);
	__klp_bsc1195950_put_shared_state();
	mutex_unlock(&module_mutex);
}

#endif /* IS_ENABLED(CONFIG_DRM_I915) */
